/*
 * CPU-specific routines originally from U-Boot.
 * See: uboot-xburst/files/arch/mips/cpu/xburst/cpu.c
 * See: u-boot/arch/mips/include/asm/cacheops.h
 *
 * Copyright (C) 2000-2009 Wolfgang Denk, DENX Software Engineering, <wd@denx.de>
 * Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "cpu.h"
#include "sdram.h"

void flush_icache_all(void)
{
	u32 addr, t = 0;

	asm volatile ("mtc0 $0, $28"); /* Clear Taglo */
	asm volatile ("mtc0 $0, $29"); /* Clear TagHi */

	for (addr = KSEG0; addr < KSEG0 + CONFIG_SYS_ICACHE_SIZE;
	     addr += CONFIG_SYS_CACHELINE_SIZE) {
		asm volatile (
			".set mips3\n\t"
			" cache %0, 0(%1)\n\t"
			".set mips2\n\t"
			:
			: "I" (Index_Store_Tag_I), "r"(addr));
	}

	/* invalicate btb */
	asm volatile (
		".set mips32\n\t"
		"mfc0 %0, $16, 7\n\t"
		"nop\n\t"
		"ori %0,2\n\t"
		"mtc0 %0, $16, 7\n\t"
		".set mips2\n\t"
		:
		: "r" (t));
}

void flush_dcache_all(void)
{
	u32 addr;

	for (addr = KSEG0; addr < KSEG0 + CONFIG_SYS_DCACHE_SIZE; 
	     addr += CONFIG_SYS_CACHELINE_SIZE) {
		asm volatile (
			".set mips3\n\t"
			" cache %0, 0(%1)\n\t"
			".set mips2\n\t"
			:
			: "I" (Index_Writeback_Inv_D), "r"(addr));
	}

	asm volatile ("sync");
}

void flush_cache_all(void)
{
	flush_dcache_all();
	flush_icache_all();
}

void handle_error_level(void)
{
	asm volatile(
		"mfc0 $t3, $12\n"	/* CP0_STATUS */
		"li $t4, 0xfffffffb\n"	/* ERL = 0 */
		"and $t3, $t3, $t4\n"
		"mtc0 $t3, $12\n"
		"nop\n");
}

void enable_interrupts(void)
{
	asm volatile(
		"mfc0 $t3, $12\n"	/* CP0_STATUS */
		"li $t4, 0x0000fc01\n"	/* IE = enable interrupts */
		"or $t3, $t3, $t4\n"
		"mtc0 $t3, $12\n"
		"nop\n");
}

void init_interrupts(void)
{
	/* Set exception registers. */

	asm volatile(
		"mtc0 $zero, $18\n"	/* CP0_WATCHLO */
		"li $t3, 0x00800000\n"	/* IV = 1 (use 0x80000200 for interrupts) */
		"mtc0 $t3, $13\n"	/* CP0_CAUSE */
		"mfc0 $t4, $12\n" 	/* CP0_STATUS */
		"li $t3, 0xffbfffff\n"	/* BEV=0 */
		"and $t3, $t3, $t4\n"
		"mtc0 $t3, $12\n"
		"nop\n");
}

void set_task(u8 asid)
{
	asm volatile(

		/* Set the ASID. */

		"mtc0 %0, $10\n"	/* CP0_ENTRYHI */
		"nop"
		:
		: "r" (asid)
		);
}

void init_stack(u32 top, u32 got, void (*function)(), u32 args[], u8 nargs)
{
	u8 i;

	asm volatile(
		"move $t3, %0\n"	/* refer to the stack frame */
		"addi $t3, $t3, -16\n"	/* refer to the first parameter */
		:
		: "r" (top)
		);

	/* Provide arguments to the function. */

	for (i = 0; i < nargs; i++)
	{
		asm volatile(
			"sw %0, 0($t3)\n"
			"addi $t3, $t3, -4\n"
			:
			: "r" (args[i])
			);
	}

	/* Store essential data for the function environment. */

	asm volatile(
		"subu %1, %1, 0x80000000\n"	/* obtain user mode addresses */
		"subu %2, %2, 0x80000000\n"
		"sw %2, -100(%0)\n"		/* store the function address as t9 */
		"sw %1, -104(%0)\n"		/* store the global pointer */
		"sw %2, -112(%0)\n"		/* store the function address as ra (for invoke_task) */
		"sw %2, -116(%0)\n"		/* store the function address as EPC (for the handler) */
		:
		: "r" (top), "r" (got), "r" (function)
		);
}

void enter_user_mode(void)
{
	asm volatile(
		"mfc0 $t3, $12\n"	/* CP0_STATUS */
		"li $t4, 0x00000010\n"	/* KSU = 2 (UM = 1) */
		"or $t3, $t3, $t4\n"
		"mtc0 $t3, $12\n"
		"nop\n");
}

void init_tlb(void)
{
	/* Wire in the kseg0 mapping and the page tables. */

	asm volatile(
		"li $t1, 1\n"		/* index of first randomly-replaced entry */
		"mtc0 $t1, $6\n"	/* CP0_WIRED */
		"mtc0 $zero, $4\n"	/* CP0_CONTEXT */
		"mtc0 $zero, $10\n"	/* CP0_ENTRYHI */
		"nop\n");
}

void map_page_index(u32 virtual, u32 physical, u32 pagesize, u8 flags, u8 asid, u32 index)
{
	u32 start = (virtual & 0xffffe000) | asid; /* VPN2 | ASID*/
	u32 lower = ((physical & 0xfffff000) >> 6) | flags;
	u32 upper = (((physical + pagesize) & 0xfffff000) >> 6) | flags;
	u32 pagemask = ((pagesize - 1) & 0xfffff000) << 1;

	asm volatile(
		"mtc0 %3, $5\n"		/* CP0_PAGEMASK */

		/* Set the index. */

		"mtc0 %4, $0\n"		/* CP0_INDEX */

		/* Set physical address. */

		"mtc0 %0, $2\n"		/* CP0_ENTRYLO0 */
		"mtc0 %1, $3\n"		/* CP0_ENTRYLO1 */

		/* Set virtual address. */

		"mtc0 %2, $10\n"	/* CP0_ENTRYHI */
		"nop\n"

		"tlbwi\n"
		"nop"
		:
		: "r" (lower), "r" (upper), "r" (start), "r" (pagemask), "r" (index)
		);
}

void init_page_table(u32 page_table, u32 virtual, u32 physical, u32 pagesize, u8 flags, u8 asid)
{
	u32 lower = ((physical & 0xfffff000) >> 6) | flags;
	u32 upper = (((physical + pagesize) & 0xfffff000) >> 6) | flags;

	/*
	With a complete address space mapping involving pairs of 4KB pages
	described by two values for each entry, there would be...

	an address space of 0x100000000 requiring...

	0x100000000 / (8 * 1024) == 0x100000000 >> 13
	== 524288 entries
	== 0x80000 entries

	Thus, each task's entries would require...

	0x80000 * 8 == 0x400000 bytes

	The kseg2 region thus permits 256 tasks occupying 0x40000000 bytes.

	However, for more modest address spaces occupying as much as 32MB there
	would be...

	an address space of 0x02000000 requiring...

	0x02000000 / (8 * 1024) == 0x02000000 >> 13
	== 4096 entries
	== 0x1000 entries

	Thus, each task's entries would only require...

	0x1000 * 8 == 0x8000 bytes
	*/

	u32 base = page_table + page_table_task_size * asid;

	/* Each page table entry corresponds to a pair of 4KB pages and holds two values. */

	u32 entry = ((virtual & 0xffffe000) >> 13) * 8;
	u32 address = base + entry;

	/* The page tables should be permanently mapped to avoid hierarchical TLB miss handling. */

	asm volatile(
		"sw %1, 0(%0)\n"
		"sw %2, 4(%0)\n"
		:
		: "r" (address), "r" (lower), "r" (upper)
		);
}

void map_page(u32 virtual, u32 physical, u32 pagesize, u8 flags, u8 asid)
{
	u32 start = (virtual & 0xffffe000) | asid; /* VPN2 | ASID*/
	u32 lower = ((physical & 0xfffff000) >> 6) | flags;
	u32 upper = (((physical + pagesize) & 0xfffff000) >> 6) | flags;
	u32 pagemask = ((pagesize - 1) & 0xfffff000) << 1;

	asm volatile(
		"mtc0 %3, $5\n"		/* CP0_PAGEMASK */

		/* Set physical address. */

		"mtc0 %0, $2\n"		/* CP0_ENTRYLO0 */
		"mtc0 %1, $3\n"		/* CP0_ENTRYLO1 */

		/* Set virtual address. */

		"mtc0 %2, $10\n"	/* CP0_ENTRYHI */
		"nop\n"

		"tlbwr\n"
		"nop"
		:
		: "r" (lower), "r" (upper), "r" (start), "r" (pagemask)
		);
}

void map_page_miss(u32 physical, u32 pagesize, u8 flags)
{
	u32 lower = ((physical & 0xfffff000) >> 6) | flags;
	u32 upper = (((physical + pagesize) & 0xfffff000) >> 6) | flags;
	u32 pagemask = ((pagesize - 1) & 0xfffff000) << 1;

	asm volatile(
		"mtc0 %2, $5\n"		/* CP0_PAGEMASK */

		/* Set physical address. */

		"mtc0 %0, $2\n"		/* CP0_ENTRYLO0 */
		"mtc0 %1, $3\n"		/* CP0_ENTRYLO1 */
		"nop\n"

		"tlbwr\n"
		"nop"
		:
		: "r" (lower), "r" (upper), "r" (pagemask)
		);
}

void unmap_page(u32 virtual, u32 physical, u32 pagesize, u8 flags, u8 asid)
{
	u32 start = (virtual & 0xffffe000) | asid; /* VPN2 | ASID*/
	u32 lower = ((physical & 0xfffff000) >> 6) | flags;
	u32 upper = (((physical + pagesize) & 0xfffff000) >> 6) | flags;
	u32 pagemask = ((pagesize - 1) & 0xfffff000) << 1;
	u32 index = 0;

	asm volatile(
		"mtc0 %4, $5\n"		/* CP0_PAGEMASK */

		/* Set physical address. */

		"mtc0 %1, $2\n"		/* CP0_ENTRYLO0 */
		"mtc0 %2, $3\n"		/* CP0_ENTRYLO1 */

		/* Set virtual address. */

		"mtc0 %3, $10\n"	/* CP0_ENTRYHI */
		"nop\n"

		/* Find an existing mapping. */

		"tlbp\n"
		"nop\n"

		/* Read the index register to see if a match was found. */

		"mfc0 %0, $0\n"		/* CP0_INDEX */
		"nop"
		: "=r" (index)
		: "r" (lower), "r" (upper), "r" (start), "r" (pagemask)
		);

	/* Return if the page is not mapped. */

	if (index & 0x80000000)
		return;

	/* Otherwise, invalidate the mapping. */

	map_page_index(virtual, physical, pagesize, flags & 0xfd, asid, index);
}
